// Harmless warning about browse information
#pragma warning (disable : 4786)
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "Compress.h"
#include "Decompress.h"

// Some test command lines
//-c C:\work\C64\MusicEditorHelp\MusicEditorHelp.prg c:\temp\t.cprg
//-co C:\work\C64\MusicEditorHelp\MusicEditorHelp.prg c:\temp\t.cprg 2
//-d c:\temp\t.cprg c:\temp\t.prg
//-co C:\work\C64\Scroller\Scroller.prg c:\temp\t.cprg 2
//-co C:\work\ReplicaNetWork\TestXPCompression\Scroller.prg c:\temp\t.cprg 2
//-co C:\work\C64\CityGame\OriginalData.prg c:\temp\t.cprg 2
//-c C:\work\C64\Decompression\TestScreen.bin c:\temp\t.cprg

u32 TestRLEPack(u8 *input,const u32 inputLength)
{
	u32 foundLeast[256];
	memset(foundLeast,0,sizeof(foundLeast));

	// Total up the occurrences of each byte value
	u32 i;
	for (i=0;i<inputLength;i++)
	{
		foundLeast[input[i]]++;
	}

	// Find the least used byte
	u8 leastIndex = 0;
	for (i=1;i<256;i++)
	{
		if (foundLeast[i] < foundLeast[leastIndex])
		{
			leastIndex = (u8) i;
		}
	}

	u8 *work = (u8*) malloc(inputLength + (inputLength / 8));
	u32 outPos = 0;
	// Save the marker
	work[outPos++] = leastIndex;

	// Now RLE pack with the least used byte as a marker
	for (i=0;i<inputLength;i)
	{
		bool match = false;
		if ( (i+3) < inputLength )
		{
			if ( (input[i] == input[i+1]) && (input[i] == input[i+2]))
			{
				match = true;
			}
		}
		// Check to see if we need to output a run
		if ( (input[i] == leastIndex) || match)
		{
			u32 j = 0;
			while ((input[i+j] == input[i]) && j < 256)
			{
				j++;
			}
			work[outPos++] = leastIndex;
			work[outPos++] = input[i];
			work[outPos++] = (u8)j-1;
			i+=j;
		}
		else
		{
			work[outPos++] = input[i++];
		}
	}

	memcpy(input,work,outPos);
	free(work);

	return outPos;
}

int main(int argc,char **argv)
{
	if (argc < 4)
	{
		printf("Compression : A very simple file compressor or decompressor suitable for C64 files.\n");
		printf("Compression -c <input file> <outfile file>\n");
		printf("Compression -d <input file> <outfile file>\n");
		printf("Use the following to skip a number of bytes from the start of the file and set a length. If the length equals 0 the length is still set from the file: -co <input file> <outfile file> <decimal offset> [decimal length]\n");
		exit(-1);
	}

	u8 input[65536];
	u8 output[73728];
	int offset = 0;
	int wantedLength = 0;

	if ((argv[1][1] == 'c') && (argv[1][2] == 'o'))
	{
		if (argc >= 5)
		{
			offset = atoi(argv[4]);
		}
		if (argc >= 6)
		{
			wantedLength = atoi(argv[5]);
		}
	}

	FILE *fp;
	fp = fopen(argv[2],"rb");
	if (!fp)
	{
		printf("Problem opening '%s' for reading\n",argv[2]);
		exit(-1);
	}

	fseek(fp,offset,SEEK_SET);

	u16 inputSize = fread(input,1,sizeof(input),fp);

	if (wantedLength)
	{
		inputSize = wantedLength;
	}

	fclose(fp);

	fp = fopen(argv[3],"wb");
	if (!fp)
	{
		printf("Problem opening '%s' for writing\n",argv[3]);
		exit(-1);
	}

	printf("Working...\n");

	if (argv[1][1] == 'c')
	{
//		inputSize = TestRLEPack(input,inputSize);
		fwrite(&inputSize,1,sizeof(inputSize),fp);

		u32 outSize;
		int bestBits = 0;
		Compression comp;
#if 1
		// For now we will always use one escape bit because it seems to consistently generate
		// the best results.
		comp.Compress(input,inputSize,output,&outSize,10,1);
#else
		comp.Compress(input,inputSize,output,&outSize,10,0);
		int i;
		for (i = 1; i < 4 ; i++)
		{
			u32 tSize;
			printf("Trying optimise %d... ",i);
			comp.Compress(input,inputSize,output,&tSize,10,i);
			if (tSize < outSize)
			{
				printf("Better!\n",i);
				bestBits = i;
				outSize = tSize;
			}
			else
			{
				printf("\n",i);
			}
		}
		comp.Compress(input,inputSize,output,&outSize,10,bestBits);
#endif

		fwrite(output,1,outSize,fp);

		printf("Input length = %x output length = %x\n",(int)inputSize,(int)outSize);
		fclose(fp);

		// Enable the following code to paranoia check the compression is OK
#if 1
		u8 temp[65536];
		u32 tempLen;
		if (Decompress(output,outSize,temp,&tempLen) || (tempLen != inputSize) || (memcmp(temp,input,tempLen) != 0))
		{
			printf("Problem during decompression\n");
			exit(-1);
		}
#endif

	}

	if (argv[1][1] == 'd')
	{
		u16 origSize = *(u16*)input;
		inputSize -= sizeof(inputSize);
		printf("Input length = %x Original output length = %x\n",(int)inputSize,(int)origSize);

		u32 outDecomp;
		if (Decompress(input + sizeof(origSize),inputSize,output,&outDecomp))
		{
			printf("Problem during decompression\n");
			exit(-1);
		}
		fwrite(output,1,outDecomp,fp);
		fclose(fp);
	}


	return 0;
}
